# =====================================================================
# csalib.awk: AWK functions for CSA.
#
# Copyright (c) 2001,2002,2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: csalib.awk,v 1.1 2003/06/10 14:45:07 carlo Exp $

# =====================================================================
# Standard C-library workalikes (not always compatible, often quite
# different from their equally-named C counterparts).
# =====================================================================

# =====================================================================
# strdup(3): string strdup(string s, int length)
# =====================================================================

function strdup(s1, len,	s2) {

   if (!len) len = 2					# see strdup(3)
   s2 = sprintf("%*s", (len + length(s1) - 1), s1)
   gsub(/ /, s1, s2)
   return s2
}

# =====================================================================
# ctime(3): string ctime(string specs)
# =====================================================================

function ctime(ts,	format, t) {

   format = "%a %b %d %H:%M:%S %Z %Y"
   if (ts == 0) CMD = "date '+" format "'"
   else
      CMD = "date -d'1970-01-01 UTC + " ts " sec' '+" format "'"

   CMD = "exec " CMD				# Spare processes.
   CMD | getline t
   close(CMD)
   return t
}

# =====================================================================
# stat(2): int stat(string file)
#
# Returns EINVAL if no arg, 0 if file exists (can be a hard/symbolic
# link) and is readable, EISDIR if file is a directory, ENOENT if file
# doesn't exist.
# =====================================================================

function stat(file,	rc) {

   if (file == "") return EINVAL
   if ((rc = (getline < file)) >= 0) { close(file); return 0 }
   if (rc == 2) return EISDIR
   else if (rc == 255) return ENOENT
   else return EIO		      # Unknown code, assume I/O error.
}

# =====================================================================
# creat(2): int creat(string file [, int flag])
#
# 'flag' can be either O_TRUNC or O_APPEND. The file will be left
# open in the specified mode on return.
# =====================================================================

function creat(file,flag,	rc) {

   if (file == "") return EINVAL
   if (flag != O_APPEND) flag = O_TRUNC

   # Check whether we can write to the requested file.
   printf("") >> file
   if ((rc=close(file)) != 0) return EACCES

   # Yes, we can.
   if (flag == O_APPEND) { printf("") >> file; return 0 }
   else printf("") > file

   return 0
}

# =====================================================================
# CSA basic functions.
# =====================================================================

# =====================================================================
# string strip(string s, [boolean middle])
#
# Strip leading and trailing blanks and TABs from the input string.
# Returns the stripped string. If 'middle' is set to '1', then multiple
# middle blanks are squeezed.
# =====================================================================

function strip(s,m) {
  sub(/^[ \t]+/,NULL,s)
  sub(/[ \t]+$/,NULL,s)
  if (m) gsub(/  +/," ",s)
  return s
}

# =====================================================================
# boolean rcassign(string name, string value [, string ofile])
#
# Turn the passed name/value pair into a valid rc(1) assignment in
# the form: name=(value_1 value_2 ... value_n). Value is split on
# octal "\001" to cope with rc(1) lists. The resulting assignment is
# then printed to 'ofile', properly escaped for rc(1).
# =====================================================================

function rcassign(name,value,ofile,		i,a,j) {

  i=j=0
  if (ofile == NULL) ofile = STDOUT

  # Help rc(1) handle non-conventional variable names.
  name = rcescape(name,1)
  if (name == "''") return EINVAL

  printf("%s=(",name) >> ofile

  # Split on "\001" to cope with rc(1) lists.

  if ((i = split(value,a,"\001"))) { 
     for (j=1; j<=i; j++) printf(" %s", rcescape(a[j],1)) >> ofile
  }

  print ")" >> ofile
}

# =====================================================================
# string rcescape(string s [, boolean set])
#
# Escape single-quotes in 'string' by doubling them as required by
# rc(1). The resulting string is then enclosed in single-quotes. If
# 'string' is null, then '()' is returned, unless the 'set' argument is
# set to '1', in which case the empty string assignment '' is returned.
# See rc(1) for more info.
# =====================================================================

function rcescape(string,set) {

  if (string == NULL) {
     if (set) return "''"
     else return "()"
  }

  gsub("'", "''", string)
  return "'" string "'"
}

# =====================================================================
# void rcmsg(string msg [, string ofile])
#
# Takes a string with all the parameters necessary for the invoking
# rc(1) script to call the underlying csaPrintMsg() library function.
# Individual parameters are extracted by splitting 'string' on octal
# value "\001".
# =====================================================================

function rcmsg(msg, ofile,		i,a) {

  if (ofile == "") ofile = STDOUT

  i = split(msg, a, "\001")

  if (a[1] !~ /^[0-9]+$/) return EINVAL

  printf("~ $CSA_WBMSG () && {CSA_WBMSG=(%s", a[1]) >> ofile

  if (a[2] != "") printf(" %s)}\n", rcescape(a[2])) >> ofile
}

# =====================================================================
# boolean isemail(string s)
#
# Checks that 'string' is a valid E-Mail address. Returns non-zero if
# it isn't. Multiple addresses can be specified in 'string', by
# saparating them with a comma ','. Blanks and/or tabs surrounding the
# address separator are silently ignored.
# =====================================================================

function isemail(string,	addr1,addr2) {

  gsub(/[ \t]*,[ \t]*/, ",", string)	# Strip empty separators.
  if(string == NULL) return EINVAL

  split(string, addr1, ",")

  for (addr2 in addr1) {
      if (addr1[addr2] !~ /^[-_=.a-zA-Z0-9]+@[-.a-zA-Z0-9]+$/)
		return EINVAL
  }
}

# =====================================================================
# string unesctbl(string s)
#
# Un-escape NoSQL special sequences \t, \n and \\
# Note: global variables, like RSTART and RLENGTH cannot be localised
# with their original names, or things won't work.
# =====================================================================

function unesctbl(s,		S,rstart,rlength) {

   while (match(s,/\\+t/)) {

      rstart=RSTART
      rlength=RLENGTH

      if (!(rlength % 2))
	 S = S substr(s,1,rstart+rlength-3) "\t"
      else
	 S = S substr(s,1,rstart+rlength-1)

      s = substr(s,rstart+rlength)
   }

   s = S s
   S = ""

   while (match(s,/\\+n/)) {

      rstart=RSTART
      rlength=RLENGTH

      if (!(rlength % 2))
	 S = S substr(s,1,rstart+rlength-3) "\n"
      else
	 S = S substr(s,1,rstart+rlength-1)

      s = substr(s,rstart+rlength)
   }

   s = S s
   S = ""

   while (match(s,/[\\][\\]/)) {

      rstart=RSTART; rlength=RLENGTH

      S = S substr(s,1,rstart)
      s = substr(s,rstart+2)
   }

   return S s
}

# =====================================================================
# string escreg(string re)
#
# Escape special characters in a string that is to be used as an AWK
# regular expression.
# =====================================================================

function escreg(re) {

  # Escape backslashes first ...
  gsub(/\\/,"\\\\\\", re)

  # ... then all other special characters.
  gsub(/\$/, "\\$", re)
  gsub(/\(/, "\\(", re)
  gsub(/\)/, "\\)", re)
  gsub(/\[/, "\\[", re)
  gsub(/\]/, "\\]", re)
  gsub(/\|/, "\\|", re)
  gsub(/\^/, "\\^", re)
  gsub(/\*/, "\\*", re)
  gsub(/\?/, "\\?", re)
  gsub(/\./, "\\.", re)

  return re
}

# =====================================================================
# string escape(string s)
#
# Escape AWK special characters ", \ and \n in a string that is to be
# eval'd again by AWK.
# =====================================================================

function escape(s) {

  # Escape backslashes first ...
  gsub(/\\/,"\\\\\\", s)

  # ... then double-quotes and newlines.
  gsub(/"/, "\\\"", s)
  gsub(/\n/, "\\n", s)

  return "\"" s "\""
}

# =====================================================================
# boolean awkassign(string name, array value \
#			[, boolean null, boolean begin, string ofile])
#
# Turn the 'value' array into a valid AWK array assignment statement.
# The No. of elements
# is expected in value[0]. The resulting assignment is then printed to
# stdout, properly escaped for AWK. Unless 'null' is non-zero, empty
# input values will result in a "delete" statement on output.
# If 'begin' is non-zero, the resulting assignment is done inside a
# "BEGIN{}" block.
# =====================================================================

function awkassign(name,value,null,begin,ofile,		i) {

  i=0
  if (ofile == NULL) ofile = STDOUT

  # Check for legal name.
  if (name !~ /^[A-Za-z][A-Za-z0-9_]*$/) return EINVAL

  if (begin) print "BEGIN {" >> ofile

  if (value[0] <= 0) {			# Handle special case.
     if (!null) printf("delete %s\n", name) >> ofile
     else printf("%s=\"\"", name) >> ofile
     return 0
  }

  printf("%s[0]=%d\n", name, value[0]) >> ofile

  while (++i <= value[0]) {
     printf("%s[%d]=%s\n", name, i, escape(value[i])) >> ofile
  }

  if (begin) print "}" >> ofile
}

# =====================================================================
# string localdate(string iso_date, [boolean res])
#
# Takes a date string, in the ISO format "YYYY-mm-dd [HH:MM:SS]" and
# returns the equivalent dd/mm/yyyy or mm/dd/yyyy string, based on
# CSA_LANG. If the optional resolution argument is "1", then HH:MM:SS
# are also returned, if they were provided by the caller.
# =====================================================================

function localdate(date,res,		local_date, token) {

  local_date = strip(date)
  if (local_date == NULL) return NULL

  if (split(date, token, /[- :]/) < 3) return NULL

  if (ENVIRON["CSA_LANG"] != "en_US") {			# Non-US
     local_date = token[3] "/" token[2] "/" token[1]
  }  else {						# US English
     local_date = token[2] "/" token[3] "/" token[1]
  }

  if (res)
     local_date = local_date " " token[4] ":" token[5] ":" token[6]

  return local_date
}

# =====================================================================
# string isodate(string local_date)
#
# Takes a date string, in the local format, and returns the corresponding
# ISO format "YYYY-mm-dd 00:00:00". No checks are done to make sure that
# the specified date is a valid one. The input date string is interpreted
# based on CSA_LANG.
# =====================================================================

function isodate(date,				iso_date, token) {

  iso_date = strip(date)
  if (iso_date == NULL) return NULL

  if (split(date, token, /[-\/ :.]/) != 3) return NULL

  if (ENVIRON["CSA_LANG"] != "en_US") {
     iso_date = token[3] "-" token[2] "-" token[1] " 00:00:00"
  }
  else {						# Default is "en_US"
     iso_date = token[3] "-" token[1] "-" token[2] " 00:00:00"
  }
  return iso_date
}

# =====================================================================
# string computerdate(string iso_date)
#
# Takes a date string, in the ISO format "YYYY-mm-dd [HH:MM:SS]" and
# returns the equivalent yyyymmdd[HHMMSS] string.
# =====================================================================

function computerdate(date) {

  gsub(/[- :]/, NULL, date)
  return date
}

# =====================================================================
# string isdate(string date, int month)
#
# Takes a date string in any format acceptable by date(1), and a month
# number, and checks whether the date value is a valid one, accounting
# for leap years, etc. Returns 0 on success, or EINVAL otherwise.
# The month No. is needed to make the check.
# Note: this function requires the GNU version of date(1).
# =====================================================================

function isdate( date, mnum,			month ) {

  CMD = "exec date --date='" date "' +%m"
  CMD | getline month
  close(CMD)

  if (month/1 == 0) return EINVAL
  if ((month/1) == (mnum/1)) return 0

  return EINVAL
}

# =====================================================================
# string adddays(string date, int days)
#
# Takes two numeric arguments in the for of 'YYYYmmdd,days', adds 'days'
# to 'yyyymmdd' and returns the result in the form of 'yyyymmdd'. If
# 'days' is a negative number then it is subtracted from 'yyyymmdd'.
# Note: this function requires the GNU version of date(1).
# =====================================================================

function adddays(date, days) {

  # Do not waste time if there is nothing to do.
  if (days == NULL) return date

  if (days > 0) {
     CMD = "exec date --date '" date " " days " day' +%Y%m%d"
  }  else {
     # Make it a positive number.
     days = days + -2 * days
     CMD = "exec date --date='" date " " days " day ago' +%Y%m%d"
  }

  CMD | getline date
  close(CMD)
  return date
}

# =====================================================================
# string uriencode(string s, boolean urp)
#
# Perform URI-encoding of special characters, according to RFC 1378. The
# set of characters that are URI-encoded is:
#
#  % # & ? + SPACE = " \t \f ' \n \r $ ( ) [ ] ` < > \ / . |
#  ; ! * { } : ~ ,
#
# Note how '%' must be escaped first, and SPACE must be done after
# '+'. See http://www.apache.org/info/css-security/ for more info on
# the so-called Cross-Site Scripting vulnerabilities. If the 'urp'
# ragument is set to '1', then the encoding is done twice for some of
# the involved characters. This meant to produce values that are to
# be included in PATH_INFO elements, rather than in the QUERY_STRING,
# as the content of PATH_INFO is unencoded a first time by the CGI
# interface itself (verified with both Apache and Boa), and then by the
# CSA decode() function. If the data contain characters such as / ? and
# &, and they are encoded only once, they will be made available to CSA
# already unencoded, and that would disrupt a number of things (for
# instance, splitting PATH_INFO on slashes would become unreliable, to
# name only one).
# =====================================================================

function uriencode(string,urp) {

   if (urp) {
      gsub(/&/, "%26", string)
      gsub(/\?/, "%3F", string)
      gsub(/=/, "%3D", string)
      gsub(/\//, "%2F", string)
   }

   # Order matters here.
   gsub(/%/, "%25", string)
   gsub(/\+/, "%2B", string)
   gsub(/  */, "+", string)

   # Order does not matter here.
   gsub(/\t/, "%09", string)
   gsub(/\f/, "%0C", string)
   gsub(/\n/, "%0A", string)
   gsub(/'/, "%27", string)
   gsub(/#/, "%23", string)
   gsub(/&/, "%26", string)
   gsub(/\?/, "%3F", string)
   gsub(/=/, "%3D", string)
   gsub(/"/, "%22", string)
   gsub(/\r/, "%0D", string)
   gsub(/\$/, "%24", string)
   gsub(/\(/, "%28", string)
   gsub(/\)/, "%29", string)
   gsub(/\[/, "%5B", string)
   gsub(/]/, "%5D", string)
   gsub(/`/, "%60", string)
   # gsub(/\001/, "%01", string)
   gsub(/</, "%3C", string)
   gsub(/>/, "%3E", string)
   gsub(/\\/, "%5C", string)
   gsub(/\//, "%2F", string)
   gsub(/\./, "%2E", string)
   gsub(/\|/, "%7C", string)
   gsub(/;/, "%3B", string)
   gsub(/!/, "%21", string)
   gsub(/\*/, "%2A", string)
   gsub(/{/, "%7B", string)
   gsub(/}/, "%7D", string)
   gsub(/:/, "%3A", string)
   gsub(/~/, "%7E", string)
   gsub(/,/, "%2C", string)

   # Add more encodings here, if necessary.

   return string
}

# =====================================================================
# string b64decode(string value)
#
# Take a base64-encoded string and decode it to its original value. 
# Being mawk(1) not 8-bit clean, the original unencode string should
# not contain values that fall outside the LOCALE charset.
# =====================================================================

function b64decode(value,		S,n,k,f,i,s,b,z,c) {

  n["/"] = 63
  n["+"] = 62

  for (k=26;k;) {
      n[sprintf(f="%c",--k+65)]=k
      if (k<10) n[sprintf(f,k+48)]=k+52
      n[sprintf(f,97+k)]=26+k
  }

  for (i=1;i<=length(value);) {
      if ((s=substr(value,i++,1)) in n){
         c=n[s]
         if (1==k%=4) {
	    b=z*4+(c-(z=c%16))/16
         }
         else if (!k) z=c
         else if (k==2) {
           b=z*16+(c-(z=c%4))/4
         }
         else if (k==3) b=z*64+c

         if (k++) S = S sprintf(f,b)
      }
  }

  return S
}

# =====================================================================
# void authset(string auth, int type [,string file)
#
# Handle basic HTTP authorization.
# =====================================================================

function authset(s,t,ofile,		a,S,err) {

   if (ofile == "") ofile = STDOUT

   if (s == "") return				# ignore null requests.

   S = s					# Copy to workarea.

   if (!sub(/^[Bb][Aa][Ss][Ii][Cc] +/,"",S)) err=1	# Basic auth only
   if (S !~ /^[a-zA-Z0-9\/+]+=*[\r]?$/) err=1		# Base64 only

   if (err) {
      # Clear all auth vars on error.
      print "HTTP_AUTHORIZATION=();CSA_AUTH_PBC=()" >> ofile
      print "CSA_AUTH_USER=();CSA_AUTH_PW=()" >> ofile
      return
   }

   split(b64decode(S),a,":")

   # For user-id and password to be valid, they must contain only
   # allowed characters. Note that the password field may be null,
   # and that this is acceptable.

   if (a[1] ~ /^[a-z][a-z][-_a-z0-9@.]*$/ && \
			a[2] ~ /^[-+_a-zA-Z0-9@.!]*$/) {

      printf("CSA_AUTH_USER='%s'\n", a[1]) >> ofile
      printf("CSA_AUTH_PBC=%s/%s\n",\
			substr(a[1],1,1),substr(a[1],2,1)) >> ofile
      rcassign("CSA_AUTH_PW", a[2], ofile)

      # The application-level token must be passed to the application,
      # to be re-sent to the client with every response if necessary.
      # Note the re-use of the same variable that would be set by the
      # server in case of HTTP authorization.

      printf("HTTP_AUTHORIZATION='%s'\n", s) >> ofile
   }

   else {
      # If anything is wrong, clear all auth vars.
      print "HTTP_AUTHORIZATION=();CSA_AUTH_PBC=()" >> ofile
      print "CSA_AUTH_USER=();CSA_AUTH_PW=()" >> ofile
   }

   # CSA_AUTH_TYPE needs to be set *anyway*, but it may be overridden
   # by server-based auth, further down.

   if (t == 0) print "CSA_AUTH_TYPE=basic" >> ofile
   else print "CSA_AUTH_TYPE=custom" >> ofile

   # Server-based authentication, if used, *MUST* override the one
   # attained by all other means, and the application-level token
   # must be cleared. If the server token does not match the allowed
   # pattern, then it is cleared too.

   if (ENVIRON["REMOTE_USER"] ~ /^[a-z][a-z][-_a-z0-9@.]*$/) {
      printf("CSA_AUTH_USER='%s'\n", ENVIRON["REMOTE_USER"]) >> ofile
      printf("CSA_AUTH_PBC=%s/%s\n",substr(ENVIRON["REMOTE_USER"],1,1),\
			substr(ENVIRON["REMOTE_USER"],2,1)) >> ofile
      print "CSA_AUTH_OK=1;CSA_AUTH_PW=();HTTP_AUTHORIZATION=()" >> ofile
      print "CSA_AUTH_TYPE=server" >> ofile
   }
   else print "REMOTE_USER=()" >> ofile
}

# =====================================================================
# string hdrtocgi(string hdr)
#
# Encode "=" and "&" characters in HTTP headers (including cookies), so
# that they can be appended to the QUERY_STRING and handled by decode()
#
# CSA HTTP custom headers (X-something) will have their names turned
# into upper-case, and with invalid characters changed into underscores
# =====================================================================

function hdrtocgi(hdr,			a,name,i,result,uc) {

   name = 1

   # remove leading blanks from names.
   sub(/^ +/, "", hdr); gsub(/; */, ";", hdr)

   i = split(hdr, a, "")
   
   # test if custom HTTP header.
   if (a[1] ~ /[Xx]/ && a[2] == "-") uc = 1
   else uc = 0

   a[0] = i

   for (i=1; i<=a[0]; i++) {

      if (a[i] == ";") {
	 # test if custom HTTP header.
	 if (a[i+1] ~ /[Xx]/ && a[i+2] == "-") uc = 1
	 else uc = 0
	 if (i < a[0]) {
	    result = result "&"
	    name = 1
	    continue
	 }
      }
      if (a[i] == "=") {
	 if (name) name = 0
	 else a[i] = "%3D"
      }
      else if (name && a[i] !~ /[A-Za-z0-9]/) a[i] = "_"
      else if (!name && a[i] == " ") a[i] = "%20"
      else if (a[i] == "&") a[i] = "%26"

      # Turn field name into upper-case, unless requested otherwise.
      if (name && uc) result = result toupper(a[i])
      else result = result a[i]
   }

   return result
}

# =====================================================================
# string xmlencode(string s)
#
# Encode special XML characters into the equivalent ISO entities. The
# set of characters that are XML-encoded is:  & < " \001
#
# Warning: '&' must be escaped first. According to the XML
# specification, characters 0 through 31 and character 127 may not
# appare anywhere in a document, not even as character references, such
# as &#31;. The only other possibility it to either remove them or turn
# them into something sensible. I took this second approach, whereby
# \001 (that is rc(1) list concatenation character) is replaced with a
# single blank. This is experimental, and may change in future versions
# of CSA.
# =====================================================================

function xmlencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand at first.
   gsub(/</, "\\&lt;", string)			# Open tag
   gsub(/"/, "\\&quot;", string)		# double quote
   gsub(/\001/, " ", string)			# rc(1)'s ^A

   return string
}

# =====================================================================
# string xmldecode(string s)
#
# Perform the inverse of xmlencode(), with the exception that the \001
# hack cannot be reverted, not being it distinquishable from all other
# blanks.
# =====================================================================

function xmldecode(string) {

   gsub(/&quot;/, "\"", string)		# double quote
   gsub(/&lt;/, "<", string)		# Open tag
   gsub(/&gt;/, ">", string)		# Close tag
					# (some programs encode it too)
   gsub(/&amp;/, "\\&", string)		# Ampersand at last.

   return string
}

# =====================================================================
# string isoencode(string s)
#
# Encode special HTML and form characters into the equivalent ISO
# entities. The set of characters that are ISO-encoded is:
#
#		& # < > | " ' ` \ \n \t \f SPACE
#
# Warning: '&' must be escaped first, then '#'. They need to be escaped
# to prevent the data from containing statements (both numeric and
# literal) that could be parsed by a Web server as valid SSI calls, or
# ampersand-escaped sequences that would act as formatting instructions
# to the Web browser. See http://www.apache.org/info/css-security/ for
# more info on the so-called Cross-Site Scripting vulnerabilities.
# =====================================================================

function isoencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand
   gsub(/#/, "\\&#35;", string)			# Hash mark
   gsub(/</, "\\&#60;", string)			# Open tag
   gsub(/>/, "\\&#62;", string)			# Close tag
   gsub(/\|/, "\\&#124;", string)		# Generic separator
   gsub(/"/, "\\&#34;", string)			# double quote
   gsub(/'/, "\\&#39;", string)			# single quote
   gsub(/`/, "\\&#96;", string)			# backtick
   # gsub(/\001/, "\\&#1;", string)		# rc(1)'s ^A
   gsub(/\n/, "\\&#10;", string)		# newline
   gsub(/\t/, "\\&#9;", string)			# tab
   gsub(/\f/, "\\&#12;", string)		# form-feed
   gsub(/ /, "\\&#32;", string)			# space
   gsub(/\\/, "\\&#92;", string)		# backslash

   return string
}

# =====================================================================
# int envtoxml([string template, string fileroot, string eval,\
#		string table, boolean trim, string tago, string tagc,\
#		string ofile])
#
# Take environment variables and use them to replace special tags in
# an XML template file. All arguments are optional, but at least the
# 'template' parameter should be passed, otherwise the result will
# hardly be of any usefulness :-)
# =====================================================================

function envtoxml(tpl,froot,eval,tbl,trim,tago,tagc,ofile,	   \
	ltago,ltagc,retago,retagc,nl,end,tlen,tag,enc,a,i,\
	input,record,j,start) {

  i=j=0

  # Check args and set defaults.
  if (tpl == "") tpl = "/dev/null"
  if (ofile == "") ofile = STDOUT
  if (froot == "") froot = ENVIRON["CSA_ROOT"] "/xml/"
  if (tago == "") tago = "$["
  if (tagc == "") tagc = "]"
  if (tago == tagc) return EINVAL # start- and end-tags must be different

  ltago = length(tago)
  ltagc = length(tagc)

  # Prepare the substitution regexp's.
  retago = escreg(tago)
  retagc = escreg(tagc)

  retago = retago "[-A-Za-z0-9_/.:]+" retagc

  # Load an extra NoSQL table if requested. This data is also loaded
  # in ENVIRON[], so it can be used to override data in there. Table
  # rows will be assigned to ENVIRON["TBL1"], ENVIRON["TBL2"], etc., with
  # ENVIRON["TBL0"] being the No. of records in table. Any table header
  # information is ignored. The names used (one lower-case letter
  # follwed by one or more numbers), are CSA-reserved names (see the docs),
  # and the "t" suffix stands for "table". Note that splitting table records
  # into fields and unescaping NoSQL special sequences is up to whoever
  # will make use of these values, often through the user-level hook
  # outvar().

  if (tbl != "") {
     ENVIRON["TBL0"] = 0
     while (getline record < tbl > 0) {
	# Store table header in a separate place.
	if (record ~ /^\001/) {
	   TBLH = record
	   continue
	}
	# Append row to ENVIRON[].
	ENVIRON["TBL" ++ENVIRON["TBL0"]] = record
     }

     close(tbl)

     tbltordf()			# prepare also the RDF version.
  }

  for (j in ENVIRON) {

      # Always skip rc(1) functions, confidential data
      # and other CSA-specific stuff.

      if (j ~ /(^fn_|^HTTP_COOKIE|PGPPASS$|^CSA_AUTH_PW|^csa0)/) {
	 ENVIRON[j] = "<!-- removed -->"
	 continue
      }


      # If CSA_AUTH_SSL is in effect, let's be careful
      # about including HTTP_AUTHORIZATION in templates.

      if (j == "HTTP_AUTHORIZATION" && ENVIRON["CSA_AUTH_SSL"]) {
	   if (ENVIRON["SSL_CIPHER_USEKEYSIZE"] == "" && \
					ENVIRON["HTTPS"] != "on") {
	      # This cannot be set to anything other than "",
	      # or there will be problems with envtoxml.awk.
	      ENVIRON[j] = ""
	      continue
	   }
      }

      # Call local output hook on non-CSA environment variables.
      if (j !~ /^[Cc][Ss][Aa]/) ENVIRON[j] = outvar(j,ENVIRON[j])

      # Include also URI-, URP-, ISO- and XML-PCDATA- encodings in
      # ENVIRON[] by default. I need to prepare all three encodings
      # because this function is used for both HTML and XML templates.
      # Note that if the string is to be included in an XML document
      # and it must represent a URI, then the the URI_ or URP_ versions,
      # whichever applies, should be used. ISO_ encoding is practically
      # only useful with HTML, and it is of no use with XML.

      ENVIRON["URI_" j] = uriencode(ENVIRON[j])
      ENVIRON["URP_" j] = uriencode(ENVIRON[j],1)
      ENVIRON["ISO_" j] = isoencode(ENVIRON[j])
      ENVIRON["XML_" j] = xmlencode(ENVIRON[j])
  }

  # Tell the local hook that defs are over.
  outvar("ENOMSG",ENOMSG)

  if (eval != "") split(eval,a,",")

  # Process the XML template file.

  while (getline input < tpl > 0) {
     nl = 1
     do {
        match(input,retago)
        if (RSTART) {
	   end = index(input, tagc)
	   tlen = end - RSTART - ltago
	   start = RSTART		# save before calling rdfprint()
	   rdfprint(substr(input, 1, start - 1),ofile)
	   tag = substr(input, start + ltago, tlen)

	   # Now handle encoding flags.
	   if (sub(/:c$/,NULL,tag)) {
	      # character-data, to be included verbatim.
	      enc = "cdata"
	   }
	   else if (sub(/:p$/,NULL,tag)) {
	      # parsed-character data.
	      enc = "pcdata"
	   }
	   else enc = "base64"			# default is binary.

	   for (i in a) {			# Is "tag" to be eval'ed ?
	       if (a[i] == tag) {
		  tag = ENVIRON[tag]		# Get tag value from env.
		  break
	       }
	   }

	   if (tag ~ /\//) {			# Catch file patterns.
	      if (tag ~ /\.\./) {		# Catch any '..' pattern.
		 rdfprint("$[" tag "]",ofile)	# Print tag literally.
	      }
	      else {
		 # Read external file, if it exists.
		 if (froot != NULL && tag !~ /^\//) {
		    tag = froot "/" tag
		    gsub(/\/\//, "/", tag)	# Turn any "//" into "/".
		 }

		 if (enc == "base64") {		# Base64.
		    system("exec mimencode -p " tag) > ofile
		 }
		 else if (enc == "cdata") {
		    # embeddable character-data, include verbatim.
		    while (getline record < tag > 0)
					rdfprint(record"\n",ofile)
		    close(tag)
		 }
		 else {
		    # parsed-character data, whatever "parsed" means in
		    # the specific context (XML in this case).

		    while (getline record < tag > 0)
				rdfprint(xmlencode(record"\n"),ofile)
		    close(tag)
		 }
		 nl = 0
	      }
	   }
	   else {
	      if (trim) ENVIRON[tag] = strip(ENVIRON[tag])
	      rdfprint(ENVIRON[tag],ofile)
	   }

	   input = substr(input, end + ltagc)
	}
	else {
	   rdfprint(input,ofile)
	   input = NULL		# Not to loop forever.
	}
     }  while (length(input))

     if (nl) rdfprint("\n",ofile)
  }

  close(tpl)
}

# =====================================================================
# string formtext(string name, string value, \
#				int size, int maxsize, boolean pass)
#
# Returns the proper HTML "text" form fragment, based on the specified
# arguments. Warning: the "value" string is not automatically
# XML-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function. This is also to
# save processing, as the caller may want to encode the text only once
# for values that are needed in several places. If the last argument
# is "0" or NULL, then a normal input field is returned, otherwise
# the result will be a "password" type input field. If size/maxsize
# are NULL or "0" then they will be ignored.
# =====================================================================

function formtext(name,value,size,max,pass,		fragment) {

  if (pass) fragment = "<input type=\"password\" name=\"" name "\""
  else fragment = "<input type=\"text\" name=\"" name "\""

  size += 0; max += 0			# Cast to numbers.

  if (size) fragment = fragment " size=\"" size "\""
  if (max) fragment = fragment " maxsize=\"" max "\""

  fragment = fragment " value=\"" value "\">"

  return fragment
}

# =====================================================================
# string formarea(string name, string value, int cols, int rows)
#
# Returns the proper HTML "textarea" form fragment, based on the
# specified arguments. Warning: the "value" string is not automatically
# XML-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function.
# =====================================================================

function formarea(name,value,cols,rows,		fragment) {

  fragment = "<textarea name=\"" name "\""

  cols += 0; rows += 0				# Cast to numbers.

  if (cols) fragment = fragment " cols=\"" cols "\""
  if (rows) fragment = fragment " rows=\"" rows "\""

  fragment = fragment ">" value "</textarea>"

  return fragment
}

# =====================================================================
# string formhidden(string name, string value)
#
# Returns the proper HTML "hidden" form fragment, based on the specified
# arguments. Warning: the "value" string is not automatically
# XML-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function.
# =====================================================================

function formhidden(name,value,		fragment) {

  fragment = "<input type=\"hidden\" name=\"" name \
	"\" value=\"" value "\">"

  return fragment
}

# =====================================================================
# string formselect(array value, array label, string name,\
#		int selected, int size, boolean multiple, string empty)
#
# Returns the proper HTML "select" form fragment, based on the specified
# arguments. If value[] and label[] contain entries that need to be
# XML-encoded, that must be done by the caller. Position "0" of either
# arrays must contain the No. of select options. If argument "empty"
# is passed, then the a trailing empty <option> element is appended to
# the resulting HTML object, with the specified label and a NULL value.
#
# Example:
#
# value[0] = 3
# value[1] = "1st cartoon"
# value[2] = "2nd cartoon"
# value[3] = "3rd cartoon"
#
# label[1] = "Donald Duck"
# label[2] = "Mickey Mouse"
# label[3] = "Uncle Scrooge"
#
# print formselect(value,label,"myselect",2)
#
#
# Result:
#
# <select name="myselect">
#  <option value="1st cartoon">Donald Duck</option>
#  <option selected value="2nd cartoon">Mickey Mouse</option>
#  <option value="3rd cartoon">Uncle Scrooge</option>
# </select>
# =====================================================================

function formselect(a1,a2,name,value,size,multi,empty,		\
					max,count,fragment) {

  count = 0

  # The No. of select entries is taken from element "0" of
  # either arrays.
  max = a1[0]
  if (!max) max = a2[0]

  value += 0; max += 0; size += 0; multi += 0	# Cast to numbers.

  if (empty != "") {
     a1[++max] = ""
     a2[max] = empty
  }

  if (!max) return NULL				# Nothing to do.

  if (name == NULL) name = "select1"		# Default name.

  fragment = "<select name=\"" name "\""

  if (size > 1) fragment = fragment " size=\"" size "\""
  if (multi) fragment = fragment " multiple"

  fragment = fragment ">\n"

  while (++count <= max) {
     fragment = fragment "<option"
     if (value == count) fragment = fragment " selected"

     fragment = fragment " value=\"" a1[count] "\">" \
		a2[count] "</option>\n"
  }

  fragment = fragment "</select>"

  return fragment
}

# =====================================================================
# string formradio(string name, string value, \
#			string label, boolean checked)
#
# Returns the proper HTML "radio" form fragment, based on the specified
# arguments. Warning: the "label" string is not automatically
# XML-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function.
# =====================================================================

function formradio(name,value,label,check,	fragment) {

  fragment = "<input type=\"radio\" name=\"" name "\""

  value += 0; check += 0			# Cast to numbers.

  if (!value) value = NULL			# "0" --> NULL

  if (check) fragment = fragment " checked"

  fragment = fragment " value=\"" value "\">" label

  return fragment
}

# =====================================================================
# void rdfprint(string s, [string ofile])
#
# Special "print" function for XML/RDF templates. Note that for this
# function to work correctly, the input template must be correctly
# formatted, that is:
#
# 1) "<?xml ....?>" in the very first line on its own. 
# 2) "<html ....>" (case insensitive) on a line of its own.
# 3) RDF begin and end tags on separate lines.
# =====================================================================

function rdfprint(s,ofile) {

  # Set defaults.
  if (ofile == "") ofile = STDOUT

  if (ENVIRON["CSA_RPC_FMT"] == "rdf") {
     if (s ~ /^[ \t]*<[Hh][Tt][Mm][Ll][^>]*>[ \t]*$/) RDFPRINT = 0
     else if (s ~ /^[ \t]*<rdf:RDF /) RDFPRINT = 1
     else if (s ~ /^[ \t]*<\/rdf:RDF>/) {
	printf("%s",s) > ofile
	RDFPRINT = 0
	return
     }
  }

  if (RDFPRINT) printf("%s", s) > ofile
}

# =====================================================================
# string rdftype(string s)
#
# Trivial RDF datatype autodetect function. Note that it is not
# absolutely bullet-proof.
# =====================================================================

function rdftype(s,		double_re,int_re,time_re,type) {

  # types auto-detection not bullet-proof, but should work in most cases.
  double_re = "^[-+]?[0-9]+\.[0-9]+$"
  int_re = "^[-+]?[0-9]+$"
  time_re = "^([1-9][0-9]-?)+T[0-2][0-9]:[0-5][0-9]:[0-5][0-9][-+]?"

  if (s != "") {
     if (s ~ double_re) type = "double"
     else if (s ~ int_re) type = "int"
     else if (s ~ time_re) type = "dateTime"
     else type = "string"
  }
  else type = "string"

  return "http://www.w3.org/2000/10/XMLSchema#" type
}

# =====================================================================
# int tbltordf(boolean trim, string toc, string body])
#
# Turn the default NoSQL "TBLx" table into a proper RDF structure.
# =====================================================================

function tbltordf(trim,toc,body,		tblh,c_names,c_values,\
						nr,nf,i) {
  # set defaults.
  if (toc == "") toc = ENVIRON["CSA_RDF_TOC"]
  if (body == "") body = ENVIRON["CSA_RDF_BODY"]

  tblh = TBLH
  if (!gsub(/\001/,"",tblh)) return 1	# missing header info.
  nf = split(tblh,c_names,"\t")

  for (nr=1; nr <= ENVIRON["TBL0"]; nr++) {	# for each record

      printf("<rdf:li rdf:resource=\"#csa-response-%d\" />\n", nr) > toc
      printf("<result rdf:ID=\"csa-response-%d\">\n", nr) > body
      printf("<value><rdf:Seq>\n") > body

      split(ENVIRON["TBL" nr ], c_values, "\t")

      for (i=1; i <= nf; i++) {		# process each field in turn.

	  printf("<rdf:li rdf:ID=\"csa-response-%d.%s\" rdf:datatype=\"",
		nr,c_names[i]) > body

	  if (trim) {
	     sub(/^[ \t]+/,"",c_values[i])
	     sub(/[ \t]+$/,c_values[i])
	  }

	  printf("%s\">%s</rdf:li>\n", rdftype(c_values[i]),\
		xmlencode(unesctbl(c_values[i]))) > body
      }

      printf("</rdf:Seq></value></result>\n") > body
  }

  # print an empty TOC if necessary.
  if (!ENVIRON["TBL0"]) printf("<rdf:li />\n") > toc

  close(toc); close(body)
}

# =====================================================================
# Main program settings.
# =====================================================================

BEGIN {

  # The CMD global variable is used to hold system commands. It is
  # a global name because AWK can have only one system command opened
  # at a time, and the current command needs to be closed before a new
  # one can be opened.

  NULL = RESULT = CMD = ""

  RDFPRINT = 1

  # Linux-style symbolic error codes (see errno(3) and <errno.h>).
  ENOENT	=   2
  EIO		=   5
  EACCES	=  13
  EISDIR	=  21
  EINVAL	=  22
  ENOMSG	=  42
  EMSGSIZE	=  90

  # Various flags.
  O_TRUNC=0; O_APPEND=1

  # Get local settings.
  CSAINST = ENVIRON["CSA_INSTALL"]
  STDOUT = ENVIRON["CSA_STDOUT"]
  STDERR = ENVIRON["CSA_STDERR"]

  # Set default values if necessary.
  if (CSAINST !~ /^\//) CSAINST = "/usr/local/csa"
  if (STDOUT !~ /^\//) STDOUT = "/dev/stdout"
  if (STDERR !~ /^\//) STDERR = "/dev/stderr"
}

# =====================================================================
# End of csalib.awk
# =====================================================================
