# =====================================================================
# xmlrpclib.awk: basic XML-RPC AWK functions for CSA.
#
# Copyright (c) 2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: xmlrpclib.awk,v 1.4 2003/05/07 09:11:26 carlo Exp $

# =====================================================================
# string xmlencode(string s)
#
# Encode special XML characters into the equivalent ISO entities. The
# set of characters that are XML-encoded is:  & < " \001
#
# Warning: '&' must be escaped first. According to the XML
# specification, characters 0 through 31 and character 127 may not
# appare anywhere in adocument, not even as character references, such
# as &#31;. The only other possibility it to either remove them or turn
# them into something sensible. I took this second approach, whereby
# \001 (that is rc(1) list concatenation character) is replaced with a
# single blank. This is experimental, and may change in future versions
# of CSA.
# =====================================================================

function xmlencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand at first.
   gsub(/</, "\\&lt;", string)			# Open tag
   gsub(/"/, "\\&quot;", string)		# double quote
   gsub(/\001/, " ", string)			# rc(1)'s ^A

   return string
}

# =====================================================================
# string xmldecode(string s)
#
# Perform the inverse of xmlencode(), with the exception that the \001
# hack cannot be reverted, not being it distinquishable from all other
# blanks.
# =====================================================================

function xmldecode(string) {

   gsub(/&quot;/, "\"", string)			# double quote
   gsub(/&lt;/, "<", string)			# Open tag
   gsub(/&amp;/, "\\&", string)			# Ampersand at last.

   return string
}

# =====================================================================
# nil httpprint(string s, [boolean len, string ofile])
#
# Special "print" function for HTTP responses, that does buffering
# and HTTP content-length calculation if necessary.
# Uses the HTTPBODY[] global array, so that its contents are preserved
# across multiple invocation of this function. The response size, if
# computed, is stored in the HTTPLEN special var, for the same reason.
# HTTPBODY[0] will contain the no. of elements in the array.
# =====================================================================

function httpprint(s,len,ofile) {

  # Set defaults.
  if (ofile == "") ofile = STDOUT

  if (len) {
     HTTPBODY[++HTTPBODY[0]] = s
     HTTPLEN += length(s)
  }
  else printf("%s", s) > ofile
}

# =====================================================================
# int envtoxml([string template, string fileroot, string eval,\
#		string table, boolean trim, string tago,\
#		string tagc, string ofile])
#
# Take environment variables and use them to replace special tags in
# an XML template file. All arguments are optional, but at least the
# 'template' parameter should be passed, otherwise the result will
# hardly be of any usefulness :-)
#
# The lightweight Boa Web Server does not set the Content-Length header
# field in HTTP replies. This is correct, IMHO, as computing such
# header may be prohibitively expensive in some cases, because we need
# to buffer the whole response either into memory or to disk, before
# such header can be computed. However, a few XML-RPC clients insist
# on pretending thast such header be present, and fail otherwise
# (http://www.dscpl.com.au/xmlrpc-debugger.php is one of those, at least
# as of February 2003). To make such clients happy I added to the CSA
# the ability to compute such header when needed.
# =====================================================================

function envtoxml(tpl,froot,eval,tbl,trim,tago,tagc,ofile,	\
	ltago,ltagc,retago,retagc,end,tlen,tag,enc,a,i,\
	in1,in2,j,clen,cmt) {

  i=j=clen=cmt=0

  # Check args and set defaults.
  if (tpl == "") tpl = "/dev/null"
  if (ofile == "") ofile = STDOUT
  if (froot == "") froot = ENVIRON["CSA_ROOT"] "/xml/"
  if (tago == "") tago = "$["
  if (tagc == "") tagc = "]"
  if (tago == tagc) return EINVAL # start- and end-tags must be different

  ltago = length(tago)
  ltagc = length(tagc)

  # Prepare the substitution regexp's.
  retago = escreg(tago)
  retagc = escreg(tagc)

  retago = retago "[-A-Za-z0-9_/.:]+" retagc

  # Load the extra table data if requested. This data is also loaded
  # in ENVIRON[], so it can be used to override data in there. The
  # relevant values will be assigned variables csat1, csat2, etc., which
  # are CSA-reserved names (see the docs), and the "t" suffix stands
  # for "table".

  if (tbl != "") {
     while (getline record < tbl > 0) {
	# Skip table header.
	if (record !~ /^\001/) j=split(record,a,"\t")
     }

     close(tbl)

     # Append to ENVIRON, in deterministic order.
     for (i=1; i<=j; i++) ENVIRON["csat" i] = a[j]
     delete a
     i = 0
  }

  for (j in ENVIRON) {

      # Always skip rc(1) functions, confidential data
      # and other CSA-specific stuff.

      if (j ~ /(^fn_|PGPPASS$|^CSA_AUTH_PW|^csa0)/) {
	 ENVIRON[j] = "<!-- removed -->"
	 continue
      }

      # If CSA_AUTH_SSL is in effect, let's be careful
      # about including REMOTE_USER in templates.

      if (j == "REMOTE_USER" && ENVIRON["CSA_AUTH_SSL"]) {
	 if (ENVIRON["SSL_CIPHER_USEKEYSIZE"] == "" && \
				ENVIRON["HTTPS"] != "on") {
	     # This cannot be set to anything other than "",
	     # or there will be problems with xmlrpctpl.awk.
	     ENVIRON[j] = ""
	     continue
	 }
      }

      # Call local output hook on variable.
      ENVIRON[j] = outvar(j,ENVIRON[j])

      # Provide also the associated XML-encoded values.
      ENVIRON["XML_" j] = xmlencode(ENVIRON[j])
  }

  if (eval != "") split(eval,a,",")

  # Clear these variables before the very first call to httpprint().
  delete HTTPBODY; HTTPLEN=0

  # Deal with HTTP headers, unless we are inside
  # a system.multicall.

  if (ENVIRON["CSA_RPC_MULTI"])
     httpprint("<value><array><data>",clen,ofile)

  else {

     # Do this only if no top-level filtering was requested.
     # See also xmlrpc-format.awk for more info.

     if (ENVIRON["CSA_RPC_FMT"] == "") {

	# Add more server types to this test if necessary.
	if (ENVIRON["SERVER_SOFTWARE"] ~ /^(Boa\/|Apache\/1\.)/) {
	   clen=1
	   printf("Content-Length: ") > ofile
	}
	else printf("\n") > ofile	# Print HTTP header separator.
     }
     else printf("\n") > ofile		# Print HTTP header separator.

     # Print standard response start if not multicall.
     httpprint("<?xml version=\"1.0\"?><methodResponse>",clen,ofile)

     if (tpl ~ /^\/?csaFault\.xml$/)
     	httpprint("<fault>",clen,ofile)
     else
     	httpprint("<params><param>",clen,ofile)
  }

  # Process the XML template file.

  while (getline in1 < tpl > 0) {

     # Deal with comments.

     # Detect continuation/end of multiline comment.
     if (cmt) {
	if (sub(/^[^<]*-->/,"",in1)) cmt=0
	else continue
     }

     # Remove inline comments.
     gsub(/<!--[^<]*-->/,"",in1)

     # Detect begin of multiline comment.
     if (sub(/^<!--[^<]*$/,"",in1)) { cmt=1; continue }

     # Restore trailing newline dropped by 'getline',
     # where appropriate:

     # detect begin of string|base64.
     if (in1 ~ /<(stri|b)[^>]+>/) nl="\n"

     # detect end of string|base64|value.
     if (in1 ~ /<\/(stri|b|v)[^>]+>/) nl=""

     in1 = in1 nl

     do {
        match(in1,retago)
        if (RSTART) {
	   end = index(in1, tagc)
	   tlen = end - RSTART - ltago
	   httpprint(substr(in1, 1, RSTART - 1),clen,ofile)
	   tag = substr(in1, RSTART + ltago, tlen)

	   # Now handle encoding flags. Only base64, XML-RPC
	   # and plain strings are supported.
	   if (sub(/:m$/,NULL,tag)) {		# Base64.
	      enc = "base64"
	   }

	   if (sub(/:X$/,NULL,tag)) {		# embeddable XML
	      enc = "xml"
	   }

	   for (i in a) {			# Is "tag" to be eval'ed ?
	       if (a[i] == tag) {
		  tag = ENVIRON[tag]		# Get tag value from env.
		  break
	       }
	   }

	   if (tag ~ /\//) {			# Catch file patterns.
	      if (tag ~ /\.\./) {		# Catch any '..' pattern.
		 # Print tag literally.
		 httpprint("$[" tag "]",clen,ofile)
	      }
	      else {
		 # Read external file, if it exists.
		 if (froot != NULL && tag !~ /^\//) {
		    tag = froot "/" tag
		    gsub(/\/\//, "/", tag)	# Turn any "//" into "/".
		 }

		 if (enc == "base64") {		# Base64.
		    CMD = "mimencode -p " tag
		    # No need to preserve newlines in base64.
		    while (CMD | getline in2 > 0)
				httpprint(in2,clen,ofile)
		    close(CMD)
		 }

		 else if (enc == "xml") {
		    # embeddable XML, include literally.
		    while (getline in2 < tag > 0)
				httpprint(in2"\n",clen,ofile)
		    close(tag)
		 }
		 else {
		    # generic text, perform XML escaping.
		    while (getline in2 < tag > 0)
			httpprint(xmlencode(in2"\n"),clen,ofile)
		    close(tag)
		 }
	      }
	   }
	   else {
	      if (trim) ENVIRON[tag] = strip(ENVIRON[tag])
	      httpprint(ENVIRON[tag],clen,ofile)
	   }

	   in1 = substr(in1, end + ltagc)
	}
	else {
	   httpprint(in1,clen,ofile)
	   in1 = NULL		# Not to loop forever.
	}
     }  while (length(in1))
  }

  close(tpl)

  if (ENVIRON["CSA_RPC_MULTI"])
     httpprint("</data></array></value>",clen,ofile)

  else {

     # Print standard response end if not multicall.

     if (tpl ~ /^\/?csaFault\.xml$/)
     	httpprint("</fault>",clen,ofile)
     else
     	httpprint("</param></params>",clen,ofile)

     if (clen) {

	HTTPBODY[++HTTPBODY[0]] = "</methodResponse>\n"
	HTTPLEN += length(HTTPBODY[HTTPBODY[0]])

	printf("%d\n\n", HTTPLEN) > ofile
	i=0
	while (++i <= HTTPBODY[0]) printf("%s", HTTPBODY[i]) > ofile

	# Free memory.
	delete HTTPBODY; HTTPLEN=0
     }
     else httpprint("</methodResponse>\n",clen,ofile)
  }
}

# =====================================================================
# End of xmlrpclib.awk
# =====================================================================
